from typing import Dict, Any
from datetime import datetime as dt
from datetime import timedelta as td
from uuid import UUID
import json

from dateparser import parse as dateparse

from metacatalog_api import models

def flatten_to_nested(flat_dict: Dict[str, str]) -> Dict[str, Any]:
    """
    Converts a flat dictionary with dot-separated keys into a nested dictionary.
    Numeric keys in dot-separated strings are treated as ordered list indices starting from 1,
    and contents within these lists are direct dictionaries without numeric keys.

    Args:
        flat_dict (Dict[str, str]): A dictionary where keys are structured as dot-separated strings.
    
    Returns:
        Dict[str, Any]: A nested dictionary representing the hierarchical structure.
    """
    nested_dict = {}

    for key, value in flat_dict.items():
        parts = key.split('.')
        current_level = nested_dict
        
        for i in range(len(parts) - 1):
            part = parts[i]

            # Check if the current part is an index and not the last part
            if part.isdigit() and (i > 0 and isinstance(current_level, list)):
                # Adjust for 1-based indexing
                index = int(part) - 1

                # Check if the list has enough elements
                while len(current_level) <= index:
                    current_level.append({})

                # Move the level to the current index
                current_level = current_level[index]
            elif i + 1 < len(parts) and parts[i + 1].isdigit():
                # If the next part is an index, ensure current level is a list
                if part not in current_level:
                    current_level[part] = []
                current_level = current_level[part]
            else:
                # Otherwise, handle as a dictionary
                if part not in current_level:
                    current_level[part] = {}
                current_level = current_level[part]

        # Set the final part's value
        last_part = parts[-1]
        if last_part.isdigit() and isinstance(current_level, list):
            # Adjust for 1-based indexing in list
            index = int(last_part) - 1
            while len(current_level) <= index:
                current_level.append(None)
            current_level[index] = value
        else:
            current_level[last_part] = value

    return nested_dict


def remove_empty_values(payload: dict) -> dict:
    """
    Removes empty values from a dictionary.
    Currently, None and '' are considered empty.
    """
    cleaned_payload = {}
    for k, v in payload.items():
        if v is None or v == '':
            continue
        elif isinstance(v, dict):
            child = remove_empty_values(v)
            if len(child) > 0:
                cleaned_payload[k] = child
        else:
            cleaned_payload[k] = v
    return cleaned_payload


def separate_string_list(string_list: str) -> list[str]:
    # separate and sanitize the string
    separated = [s.strip() for s in string_list.split(',')]
    sanitized = [chunk for chunk in separated if chunk != '']

    return sanitized

def dict_to_pg_payload(payload: dict) -> dict:
    """

    """
    insert_payload = {k: v for k, v in payload.items()}
    for k, v in insert_payload.items():
        # handle NULL
        if v is None or v == 'None':
            insert_payload[k] = 'NULL'
        # handle Array types
        elif isinstance(v, list) and all(isinstance(i, str) for i in v):
            insert_payload[k] = "'{" + ','.join(v) + "}'"
        # handle strings
        elif isinstance(v, str):
            insert_payload[k] = f"'{v}'"
        elif isinstance(v, dict) and len(v) == 0:
            insert_payload[k] = "'{}'::jsonb"
        elif isinstance(v, dict):
            insert_payload[k] = dict_to_pg_payload(v)
        elif isinstance(v, dt):
            insert_payload[k] = f"'{v.isoformat()}'"
        elif isinstance(v, UUID):
            insert_payload[k] = f"'{v}'"
    
    return insert_payload


def metadata_payload_to_model(payload: dict) -> models.MetadataPayload:
    """
    Converts a payload dictionary to a Metadata model.
    """
    # create the payload
    payload = flatten_to_nested(payload)
    
    # extract the license
    license = models.License(
        id=payload['license_id'],
        by_attribution=payload['license'].get('by_attribution'),
        share_alike=payload['license'].get('share_alike'), 
        commercial_use=payload['license'].get('commercial_use'), 
        short_title=payload['license']['short_title'],
        title=payload['license']['title'],
        summary=payload['license']['summary'],
        link=payload['license']['link'],
    )

    # extract the variable
    variable = models.Variable(**payload.pop('variable'))

    # extract the first author
    author = models.Author(**payload['authors'][0])
    authors = [models.Author(**a) for a in payload['authors'][1:]]

    # extract the location
         # extract the location
    if 'location' in payload:
        # Validate coordinates exist and are numeric
        loc = payload['location']
        if not all(k in loc for k in ('lon', 'lat')):
            raise ValueError("Location must contain 'lon' and 'lat' coordinates")
        try:
            lon = float(loc['lon'])
            lat = float(loc['lat'])
            # Basic coordinate validation
            if not (-180 <= lon <= 180 and -90 <= lat <= 90):
                raise ValueError("Invalid coordinate values")
        except (ValueError, TypeError):
            raise ValueError("Coordinates must be valid numbers")
        # Use parameterized format to prevent SQL injection
        location = f"POINT ({lon:f} {lat:f})"
    else:
        location = 'NULL'   

    meta = models.MetadataPayload(
        title=payload['title'],
        abstract=payload['abstract'],
        external_id=payload.get('external_id'),
        version=1,
        license=license,
        is_partial=False,
        comment=payload.get('comment'),
        location=location,
        variable=variable,
        citation=payload.get('citation'),
        embargo=False,
        embargo_end=dt.now() + td(days=2*365),
        publication=dt.now(),
        lastUpdate=dt.now(),
        author=author,
        authors=authors,
        details=[models.Detail(**d) for d in payload['details']],
    )

    return meta


def datasource_payload_to_model(payload: dict) -> models.DataSource:
    # create the payload
    payload = flatten_to_nested(payload)

    # remove all empty values
    payload = remove_empty_values(payload)

    # check if a temporal scale is provided
    if 'temporal_scale' in payload:
        temporal_scale = models.TemporalScale(
            resolution=payload['temporal_scale']['resolution'],
            extent=[
                dateparse(payload['temporal_scale']['observation_start']),
                dateparse(payload['temporal_scale']['observation_end'])
            ],
            support=1.0,
            dimension_names=separate_string_list(payload['temporal_scale']['dimension_names'])
        )
    else:
        temporal_scale = None
    
    # check if a spatial scale is provided
    if 'spatial_scale' in payload:
        spatial_scale = models.SpatialScale(
            resolution=payload['spatial_scale']['resolution'],
            extent=payload['spatial_scale']['extent'],
            support=1.0,
            dimension_names=separate_string_list(payload['spatial_scale']['dimension_names'])
        )
    else:
        spatial_scale = None

    # check if the datasource type is provided
    dtype = models.DataSourceType(**payload['type'])

    # handle the args
    args = payload.get('args', {})
    if not isinstance(args, dict):
        args = dict(__literal__=args)
        args = json.dumps(args)

    # finally create the datasource
    datasource = models.DataSource(
        path=payload['path'],
        type=dtype,
        variable_names=separate_string_list(payload['variable_names']),
        temporal_scale=temporal_scale,
        spatial_scale=spatial_scale,
        args=args,
        encoding=payload.get('encoding', 'utf-8')
    )

    return datasource



"""
This module contains extra views that are only helpful if running the HTML version
of Metacatalog API. These views are rendered as a FastAPI blueprint and can be loaded
by any main template to make editing easier for the user.
"""
from pathlib import Path

from fastapi import APIRouter, Request
from fastapi.templating import Jinja2Templates

# initialize the router
edit_router = APIRouter()

# initialize the templates
templates = Jinja2Templates(directory=Path(__file__).parent / "templates")


@edit_router.get("/leaflet_draw.html")
def leaflet_draw(request: Request, geom: str = 'marker'):
    if geom.lower() == 'marker':
        return templates.TemplateResponse(request=request, name="leaflet_marker.html", context={})
    elif geom.lower() == 'extent':
        return templates.TemplateResponse(request=request, name="leaflet_extent.html", context={})


"""
The combined module is a Pydantic model of the entire datamodel.

This is considered an intermediate step to move from pydantic and SQLAlchemy models to
SQLModel in a future version. You can use this model to serialize and deserialize the
the metacata without the PostgreSQL database from JSON dumps.

"""
from typing import Optional,  Dict, List, Tuple
from datetime import datetime, timedelta
from enum import Enum
import json

from pydantic import BaseModel
from pydantic import UUID4, computed_field


class Unit(BaseModel):
    id: int
    name: str
    symbol: str

class Thesaurus(BaseModel):
    id: int
    uuid: str
    name: str
    title: str
    organisation: str
    url: str
    description: str

class Keyword(BaseModel):
    id: int
    uuid: str
    value: str
    path: str
    thesaurusName: Thesaurus

class Variable(BaseModel):
    id: int
    name: str
    symbol: str
    unit: Unit
    column_names: List[str]
    keyword: Optional[Keyword] = None   

class License(BaseModel):
    id: int
    short_title: str
    title: str
    by_attribution: Optional[bool] = None
    share_alike: Optional[bool] = None
    commercial_use: Optional[bool] = None
    summary: str
    link: str

class Detail(BaseModel):
    id: Optional[int] = None 
    key: Optional[str] = None
    stem: Optional[str] = None
    value: Optional[str | float | bool | dict] = None

class DataSourceNames(Enum):
    internal = 'internal'
    external = 'external'
    csv = 'csv'
    local = 'local'
    netCDF = 'netCDF'

class DataSourceType(BaseModel):
    id: Optional[int] = None
    #name: DataSourceNames
    name: str
    title: str
    description: Optional[str] = None

class TemporalScale(BaseModel):
    id: Optional[int] = None
    dimension_names: List[str] = []
    extent: Tuple[datetime, datetime]
    resolution: timedelta
    support: float

    @computed_field
    def support_iso(self) -> timedelta:
        return self.resolution * self.support
    
    @computed_field
    def observation_start_iso(self) -> datetime:
        return self.extent[0]
    
    @computed_field
    def observation_end_iso(self) -> datetime:
        return self.extent[1]
 
class SpatialScale(BaseModel):
    id: Optional[int] = None
    dimension_names: List[str] = []
    extent: Optional[str] = None
    resolution: Optional[int] = None
    resolution_str: Optional[str] = None
    support: Optional[float] = None
    support_str: Optional[str] = None

class DataSource(BaseModel):
    id: Optional[int] = None
    path: str
    type: Optional[DataSourceType] = None
    variable_names: List[str] = []
    temporal_scale: Optional[TemporalScale] = None
    spatial_scale: Optional[SpatialScale] = None
    args: Dict = {}
    encoding: Optional[str] = 'utf-8'

    class Config:
        use_enum_values = True

class Author(BaseModel):
    id: Optional[int] = None
    uuid: Optional[UUID4] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    is_organisation: Optional[bool] = None
    affiliation: Optional[str] = None
    attribution: Optional[str] = None
    organisation_name: Optional[str] = None
    organisation_abbrev: Optional[str] = None

class Location(BaseModel):
    wkt: str

    @property
    def geojson(self):
        return 
    
    @geojson.setter
    def geojson(self, value):
        pass

class Metadata(BaseModel):
    id: int
    uuid: UUID4
    title: str
    abstract: str
    external_id: Optional[str] = None
    location: Optional[str] = None
    locationShape: Optional[str] = None
    version: int = 1
    license: License
    latest_version_id: Optional[int] = None
    is_partial: bool = False
    comment: Optional[str] = None
    citation: Optional[str] = None
    embargo: bool = False
    embargo_end: Optional[datetime] = None
    publication: Optional[datetime] = None
    lastUpdate: Optional[datetime] = None

    author: Author
    authors: List[Author] = []
    variable: Variable
    datasource: Optional[DataSource] = None
    details: List[Detail] = []
    keywords: List[Keyword] = []


class MetadataPayload(Metadata):
    id: Optional[int] = None
    uuid: Optional[str] = None


def from_file(file_path: str) -> Metadata:
    """
    Load a metadata file from a file path and return the Metadata model.
    """
    # open the file and parse
    with open(file_path, 'rb') as f:
        json_data = json.load(f)
    
    return Metadata(**json_data)



from typing import Literal, Optional, List
from pathlib import Path
from contextlib import asynccontextmanager
import logging

from fastapi import FastAPI, HTTPException, Request
from fastapi.templating import Jinja2Templates
from fastapi.responses import HTMLResponse
from pydantic_geojson import FeatureCollectionModel

from metacatalog_api import core
from metacatalog_api.views import editing_tools
from metacatalog_api import  utils

logger = logging.getLogger('uvicorn.error')

# before we initialize the app, we check that the database is installed and up to date
@asynccontextmanager
async def lifespan(app: FastAPI):
    # check if the entries table can be found in the database
    with core.connect() as session:
        if not core.db.check_installed(session):
            logger.info("Database not installed, installing...")
            core.db.install(session, populate_defaults=True)
            logger.info("Database installed.")

    # now we yield the application
    yield

    # here we can app tear down code - i.e. a log message

# define the format literal
FMT = Optional[Literal['html', 'json']]


# build the base app
app = FastAPI(lifespan=lifespan)

# add the templates
templates = Jinja2Templates(directory=Path(__file__).parent / "templates")


@app.get('/')
def index(request: Request):
    return templates.TemplateResponse(request=request, name="index.html", context={})


@app.get('/entries')
@app.get('/entries.{fmt}')
def get_entries(request: Request, fmt: FMT = None, offset: int = 0, limit: int = 100, search: str = None, title: str = None, description: str = None, variable: str = None):
    """
    Load all entries from the database. 
    You may add a filter, limit and offset.
    """
    # build the filter
    filter = {'entries.title': title, 'entries.abstract': description, 'variables.name': variable}

    # sanitize the search
    if search is not None and search.strip() == '':
        search = None

    # call the function
    entries = core.entries(offset, limit, search=search,  filter={k: v for k, v in filter.items() if v is not None}) 
    
    # check if we should return html
    if fmt == 'html':
        return templates.TemplateResponse(request=request, name="entries.html", context={"entries": entries})
    return entries


@app.get('/entries/new.html')
def new_entry_page(request: Request):
    return templates.TemplateResponse(request=request, name="add_entry.html", context={})


@app.post('/entries')
async def insert_entry(request: Request):
    # collect form data an convert it to a dict
    form = await request.form()
    data = dict(form)

    # check if there is a variable id to be loaded
    if 'variable_id' in data or 'variable.id' in data:
        vid = data.pop('variable_id', data.pop('variable.id'))
        variable = core.variables(id=vid)[0]

        # append the variable to the data
        data['variable'] = variable.model_dump()

    # second: turn the form data into a metadata payload
    payload = utils.metadata_payload_to_model(data)

    # check if the payload contained a datasource
    if 'datasource' in payload:
        datasource = utils.datasource_payload_to_model(payload['datasource'])
        payload['datasource'] = datasource
    
    # add the entry and return the response
    entry = core.add_entry(payload)
    
    return HTMLResponse(f'<pre><code>{entry}</code></pre>')

@app.get('/entries/{id}')
@app.get('/entries/{id}.{fmt}')
def get_entry(id: int, request: Request, fmt: Literal['xml'] | FMT = None):
    # call the function
    entries = core.entries(ids=id)
    
    if len(entries) == 0:
        return HTTPException(status_code=404, detail=f"Entry of <ID={id}> not found")
    
    # check if we should return html
    if fmt == 'html':
        return templates.TemplateResponse(request=request, name="entry.html", context={"entry": entries[0]})
    elif fmt == 'xml':
        return templates.TemplateResponse(request=request, name="entry.xml", context={"entry": entries[0]}, media_type='application/xml')
    return entries[0]


@app.get('/locations.{fmt}', response_model=FeatureCollectionModel)
def get_entries_geojson(request: Request, fmt: FMT = None, search: str = None, offset: int = None, limit: int = None, ids: int | List[int] = None):
    # check if we should return html
    if fmt == 'html':
        return templates.TemplateResponse(request=request, name="map.html", context={})
    
    # in all other casese call the function and return the feature collection
    geometries = core.entries_locations(limit=limit, offset=offset, search=search, ids=ids)
    
    return geometries


@app.get('/licenses.{fmt}')
def get_licenses(request: Request, fmt: FMT, license_id: Optional[int] = None, variant: Literal['list', 'select'] = 'select'):
    # call the function
    try:
        licenses = core.licenses(id=license_id)
    except Exception as e:
        return HTTPException(status_code=404, detail=str(e))

    # check if we should return html
    if fmt == 'html':
        # check the number if a id was given
        if license_id is not None:
            return templates.TemplateResponse(request=request, name="license.html", context={"license": licenses.model_dump()})
        else:
            return templates.TemplateResponse(request=request, name="licenses.html", context={"licenses": licenses, "variant": variant})
    else:
        return licenses


@app.get('/authors/new.html')
def new_author(request: Request):
    return templates.TemplateResponse(request=request, name="author.html", context={})

@app.get('/authors')
@app.get('/authors.{fmt}')
@app.get('/entries/{entry_id}/authors.{fmt}')
def get_authors(request: Request, fmt: FMT = None, entry_id: int = None, author_id: int = None, search: str = None, exclude_ids: List[int] = None, target: str = None, offset: int = None, limit: int = None):
    try:
        authors = core.authors(id=author_id, entry_id=entry_id, search=search, exclude_ids=exclude_ids, offset=offset, limit=limit)
    except Exception as e:
        return HTTPException(status_code=404, detail=str(e))

    if fmt == 'html':
        if author_id is not None:
            return templates.TemplateResponse(
                request=request,
                name="author.html",
                context={"author": authors, 'variant': 'fixed', 'target': target}
            )
        return templates.TemplateResponse(
            request=request, 
            name="authors.html", 
            context={"authors": authors, 'variant': 'select' if entry_id is None else 'list', 'target': target}
        )
    else:
        return [author.model_dump() for author in authors]


@app.get('/authors/{author_id}')
@app.get('/authors/{author_id}.json')
def get_author(author_id: int, request: Request):
    try:
        author = core.authors(id=author_id)
        return author
    except Exception as e:
        return HTTPException(status_code=404, detail=str(e))


@app.get('/variables')
@app.get('/variables.{fmt}')
def get_variables(request: Request, fmt: FMT = None, offset: int = None, limit: int = None):
    if fmt == 'html':
        return templates.TemplateResponse(request=request, name="variables.html", context={})
    try:
        variables = core.variables(only_available=False, offset=offset, limit=limit)
    except Exception as e:
        return HTTPException(status_code=404, detail=str(e))
    
    if fmt == 'html':
        return templates.TemplateResponse(
            request=request, 
            name="variables.html", 
            context={"variables": variables}
        )
    else:
        return variables

@app.get('/details/new.html')
def new_details(request: Request):
    return templates.TemplateResponse(request=request, name="details.html", context={})

@app.get('/datasources/new.html')
def new_datasource(request: Request):
    # load the datasource types
    types = core.datatypes()
    return templates.TemplateResponse(request=request, name="add_datasource.html", context={"types": types})

@app.post('/datasources')
async def add_datasource(request: Request):
    # collect form data an convert it to a dict
    form = await request.form()
    data = dict(form)

    # check that the type is provided
    if 'type_id' in data:
        dtype = core.datatypes(id=data['type_id'])[0] 
        data['type'] = dtype.model_dump()

    # turn this into a payload
    payload = utils.datasource_payload_to_model(data)

    # add the datasource and return the response
    datasource = core.add_datasource(-1, payload)

    # add the entry and return the response
    return HTMLResponse(f'<pre><code>{datasource}</code></pre>')

# register the editing tools
app.include_router(editing_tools.edit_router, prefix='/utils')

if __name__ == "__main__":
    import uvicorn
    uvicorn.run('server:app', host="0.0.0.0", port=8000, reload=True)



from typing import List, Dict
from pathlib import Path
from uuid import uuid4
import json

from psycopg import Cursor
from pydantic_geojson import FeatureCollectionModel

from metacatalog_api.models import Author, Metadata, MetadataPayload, License, Variable, DataSource, DataSourceType
from metacatalog_api import utils


SQL_DIR = Path(__file__).parent / "sql"

# helper function to load sql files
def load_sql(file_name: str) -> str:
    path = Path(file_name)
    if not path.exists():
        path = SQL_DIR / file_name
    
    with open(path, 'r') as f:
        return f.read()


def install(cursor: Cursor, schema: str = 'public', populate_defaults: bool = True) -> None:
    # get the install script
    install_sql = load_sql(SQL_DIR / 'maintain' /'install.sql').format(schema=schema)

    # execute the install script
    cursor.execute(install_sql)

    # populate the defaults
    if populate_defaults:
        pupulate_sql = load_sql(SQL_DIR / 'maintain' / 'defaults.sql').replace('{schema}', schema)
        cursor.execute(pupulate_sql)


def check_installed(cursor: Cursor, schema: str = 'public') -> bool:
    try:
        info = cursor.execute(f"SELECT * FROM information_schema.tables WHERE table_schema = '{schema}' AND table_name = 'entries'").fetchone()
        return info is not None
    except Exception:
        return False
    

def get_entries(session: Cursor, limit: int = None, offset: int = None, filter: Dict[str, str] = {}) -> List[Metadata]:
    # build the filter
    if len(filter.keys()) > 0:
        expr = []
        # handle whitespaces
        for col, val in filter.items():
            if '*' in val:
                val = val.replace('*', '%')
            if '%' in val:
                expr.append(f"{col}  LIKE '{val}'")
            else:
                expr.append(f"{col}='{val}'")
        # build the filter
        filt = " WHERE " + " AND ".join(expr)
    else:
        filt = ""
    
    # handle offset and limit
    lim = f" LIMIT {limit} " if limit is not None else ""
    off = f" OFFSET {offset} " if offset is not None else ""

    # get the sql for the query
    sql = load_sql("get_entries.sql").format(filter=filt, limit=lim, offset=off)

    # execute the query
    results = [r for r in session.execute(sql).fetchall()]

    return [Metadata(**result) for result in results]


def get_entries_by_id(session: Cursor, entry_ids: int | List[int], limit: int = None, offset: int = None) -> List[Metadata]:
    if isinstance(entry_ids, int):
        entry_ids = [entry_ids]

    # build the filter
    filt = f"WHERE entries.id IN ({', '.join([str(e) for e in entry_ids])})"

    # handle offset and limit
    lim = f" LIMIT {limit} " if limit is not None else ""
    off = f" OFFSET {offset} " if offset is not None else ""
    
    # get the sql for the query
    sql = load_sql("get_entries.sql").format(filter=filt, limit=lim, offset=off)

    # execute the query
    results = [r for r in session.execute(sql).fetchall()]

    return [Metadata(**result) for result in results]


def get_entries_locations(session: Cursor, ids: List[int] = None, limit: int = None, offset: int = None) -> FeatureCollectionModel:
    # build the id filter
    if ids is None or len(ids) == 0:
        filt = ""
    else:
        filt = f" AND entries.id IN ({', '.join([str(i) for i in ids])})"
    
    # build limit and offset
    lim = f" LIMIT {limit} " if limit is not None else ""
    off = f" OFFSET {offset} " if offset is not None else ""

    # load the query
    sql = load_sql("entries_locations.sql").format(filter=filt, limit=lim, offset=off)

    # execute the query
    result = session.execute(sql).fetchone()['json_build_object']
        
    if result['features'] is None:
        return dict(type="FeatureCollection", features=[])
    
    return result
    

class SearchResult:
    id: int
    matches: List[str]
    weights: int


def search_entries(session: Cursor, search: str, limit: int = None, offset: int = None) -> List[SearchResult]:
    # build the limit and offset
    lim = f" LIMIT {limit} " if limit is not None else ""
    off = f" OFFSET {offset} " if offset is not None else ""

    # get the sql for the query
    sql = load_sql("search_entries.sql").format(prompt=search, limit=lim, offset=off)

    # execute the query
    search_results = [r['search_meta'] for r in session.execute(sql).fetchall()]

    return search_results


def get_authors(session: Cursor, search: str = None, exclude_ids: List[int] = None, limit: int = None, offset: int = None) -> List[Author]:
    # build the filter
    filt = "" if search is None and exclude_ids is None else "WHERE "
    if search is not None:
        filt = f" last_name LIKE {search} OR first_name LIKE {search} OR organisation_name LIKE {search} "
    if exclude_ids is not None:
        if filt != "":
            filt += " AND "
        filt += f" id NOT IN ({', '.join([str(i) for i in exclude_ids])})"
    
    # handle limit and offset
    lim = f" LIMIT {limit} " if limit is not None else ""
    off = f" OFFSET {offset} " if offset is not None else ""
    
    # build the basic query
    sql = "SELECT * FROM persons {filter} {offset} {limit};".format(filter=filt, offset=off, limit=lim)

    # execute the query
    results = session.execute(sql).fetchall()

    return [Author(**result) for result in results]

def get_authors_by_entry(session: Cursor, entry_id: int) -> List[Author]:
    # build the query
    sql = load_sql("get_authors_by_entry.sql").format(entry_id=entry_id)

    # execute the query
    results = session.execute(sql).fetchall()

    return [Author(**result) for result in results]

def get_author_by_id(session: Cursor, id: int) -> Author:
    # build the sql
    sql = "SELECT * FROM persons WHERE id={id};".format(id=id)

    # execute the query
    author = session.execute(sql).fetchone()

    if author is None:
        return None
    else:
        return Author(**author)

def get_variables(session: Cursor, limit: int = None, offset: int = None) -> List[Variable]:
    # build the filter
    filt = ""
    off = f" OFFSET {offset} " if offset is not None else ""
    lim = f" LIMIT {limit} " if limit is not None else ""
    
    # build the basic query
    sql = load_sql('get_variables.sql').format(filter=filt, offset=off, limit=lim)

    # execute the query
    results = session.execute(sql).fetchall()
    
    # build the model
    try:
        variables = [Variable(**result) for result in results]
    except Exception as e:
        print(e)
        raise e
    
    return variables
    
def get_available_variables(session: Cursor, limit: int = None, offset: int = None) -> List[Variable]:
    # build the filter
    filt = ""
    off = f" OFFSET {offset} " if offset is not None else ""
    lim = f" LIMIT {limit} " if limit is not None else ""

    # build the basic query
    sql = load_sql('get_available_variables.sql').format(filter=filt, offset=off, limit=lim)

    # execute the query
    results = session.execute(sql).fetchall()

    return [Variable(**result) for result in results]

def get_variable_by_id(session: Cursor, id: int) -> Variable:
    # build the filter
    filt = f" WHERE v.id={id}"
    off = ""
    lim = ""

    # build the basic query
    sql = load_sql('get_variables.sql').format(filter=filt, offset=off, limit=lim)

    # execute the query
    result = session.execute(sql).fetchone()

    if result is None:
        raise ValueError(f"Variable with id {id} not found")
    else:
        return Variable(**result)

def get_licenses(session: Cursor, limit: int = None, offset: int = None) -> List[License]:
    # build the filter
    filt = ""
    off = f" OFFSET {offset} " if offset is not None else ""
    lim = f" LIMIT {limit} " if limit is not None else ""
    
    # build the basic query
    sql = load_sql('get_licenses.sql').format(filter=filt, limit=lim, offset=off)

    # execute the query
    results = session.execute(sql).fetchall()

    return [License(**result) for result in results]

def get_license_by_id(session: Cursor, id: int) -> License:
    # build the filter
    filt = f" WHERE id={id}"

    # build the basic query
    sql = load_sql('get_licenses.sql').format(filter=filt, offset='', limit='')

    # execute the query
    result = session.execute(sql).fetchone()

    if result is None:
        raise ValueError(f"License with id {id} not found")
    else:
        return License(**result)

def get_entry_authors(session: Cursor, entry_id: int) -> List[Author]:
    # build the query
    sql = load_sql("get_entry_authors.sql").format(entry_id=entry_id)

    # execute the query
    results = session.execute(sql).all()

    return [Author(**result) for result in results]

def get_datatypes(session: Cursor, id: int = None) -> List[DataSourceType]:
    # build the query
    sql = "SELECT * FROM datasource_types"

    # handle the id
    if id is not None:
        sql += f" WHERE id={id}"
    sql += ";"

    # execute the query
    results = session.execute(sql).fetchall()

    return [DataSourceType(**result) for result in results]


def get_datasource_by_id(session: Cursor, id: int) -> DataSource:
    # handle the filter
    raise NotImplementedError


def add_entry(session: Cursor, payload: MetadataPayload) -> Metadata:
   # execute the query
    insert_payload = utils.dict_to_pg_payload(payload.model_dump())

    # fill out a few fields
    # get the sql for inserting a new entry
    sql = load_sql('insert_entry.sql').format(**insert_payload)

    # execute the query
    entry_id = session.execute(sql).fetchone()['id']

    # add the first author
    insert_author = load_sql('insert_author_to_entry.sql')
    
    # add a uuid if there is none
    if payload.author.uuid is None:
        payload.author.uuid = uuid4()
    author_payload = utils.dict_to_pg_payload(payload.author.model_dump())
    session.execute(insert_author.format(entry_id=entry_id, role="'author'", order=1, **author_payload))

    # add co-authors
    if payload.authors is not None and len(payload.authors) > 1:
        for author, idx in enumerate(payload.authors[1:]):
            if author.uuid is None:
                author.uuid = uuid4()
            author_payload = utils.dict_to_pg_payload(author.model_dump())
            session.execute(insert_author.format(entry_id=entry_id, role="'coAuthor'", order=idx + 2, **author_payload))

    # add the details
    if payload.details is not None and len(payload.details) > 0:
        detail_sql = load_sql('insert_detail_to_entry.sql')
        for detail in payload.details:
            # handle the data type
            # TODO: here we remove the decrecated fields 'title' and 'description' - this can be cleaned up when we update the database
            detail_payload = utils.dict_to_pg_payload({k: i for k, i in detail.model_dump().items() if k not in  ['value', 'title', 'descripttion']})
            
            # handle literals
            if isinstance(detail.value, dict):
                detail_payload['raw_value'] = detail.value
            else:
                detail_payload['raw_value'] = "'" + json.dumps({"__literal__": detail.value}) + "'"
            session.execute(detail_sql.format(entry_id=entry_id, **detail_payload))

    # TODO: add the keywords
    
    # load the entry from the database
    entry = get_entries_by_id(session, entry_ids=entry_id)[0]
    return entry

def add_datasource(session: Cursor, entry_id: int, datasource: DataSource) -> int:
    # get the insert payload
    insert_payload = utils.dict_to_pg_payload(datasource.model_dump())
    
    # we have to handle all optional settings to avoid errors on formatting the SQL
    if 'temporal_scale' not in insert_payload or insert_payload['temporal_scale'] == 'NULL':
        insert_payload['temporal_scale'] = {
            'resolution': 'NULL',
            'observation_start': 'NULL',
            'observation_end': 'NULL',
            'support': 'NULL',
            'dimension_names': 'NULL'
        }
    else:
        insert_payload['temporal_scale'] = {
            'resolution': f"'{insert_payload['temporal_scale']['resolution']}'",
            'observation_start': f"'{insert_payload['temporal_scale']['extent'][0]}'", 
            'observation_end': f"'{insert_payload['temporal_scale']['extent'][1]}'",
            'support': insert_payload['temporal_scale']['support'],
            'dimension_names': insert_payload['temporal_scale']['dimension_names']
            }
    if 'spatial_scale' not in insert_payload or insert_payload['spatial_scale'] == 'NULL':
        insert_payload['spatial_scale'] = {
            'resolution': 'NULL',
            'extent': 'NULL',
            'support': 'NULL',
            'dimension_names': 'NULL'
        }

    # get the sql for inserting a new datasource
    sql = load_sql('insert_datasource.sql').format(**insert_payload, entry_id=entry_id)

    # execute the query
    datasource_id = session.execute(sql).fetchone()
    
    # return datasource
    return datasource_id['datasource_id']




from typing import List
import os
from pathlib import Path
from contextlib import contextmanager

import psycopg
import psycopg.rows
from metacatalog_api import models
from dotenv import load_dotenv
from pydantic_geojson import FeatureCollectionModel

from metacatalog_api import db


load_dotenv()

METACATALOG_URI = os.environ.get("METACATALOG_URI", 'postgresql://metacatalog:metacatalog@localhost:5432/metacatalog')
SQL_DIR = Path(__file__).parent / "sql"

#METACATALOG_URI="postgresql://postgres:postgres@localhost:5433/metacatalog"
print(METACATALOG_URI)

@contextmanager
def connect(autocommit: bool = True):
    with psycopg.connect(METACATALOG_URI, autocommit=autocommit) as con:
        with con.cursor(row_factory=psycopg.rows.dict_row) as cur:
            yield cur


def entries(offset: int = 0, limit: int = 100, ids: int | List[int] = None,  search: str = None, filter: dict = {}) -> list[models.Metadata]:
    # check if we filter or search
    with connect() as session:
        if search is not None:
            search_results = db.search_entries(session, search, limit=limit, offset=offset)

            if len(search_results) == 0:
                return []
            # in any other case get them by id
            # in any other case, request the entries by id
            results = db.get_entries_by_id(session=session, entry_ids=[r["id"] for r in search_results])

            return results
        elif ids is not None:
            results = db.get_entries_by_id(session, ids, limit=limit, offset=offset)
        else:
            results = db.get_entries(session, limit=limit, offset=offset, filter=filter)

    return results


def entries_locations(ids: int | List[int] = None, limit: int = None, offset: int = None, search: str = None, filter: dict = {}) -> FeatureCollectionModel:
    # handle the ids
    if ids is None:
        ids = []
    if isinstance(ids, int):
        ids = [ids]
    
    # check if we filter or search
    with connect() as session:
        # run the search to ge the ids
        if search is not None:
            search_results = db.search_entries(session, search, limit=limit, offset=offset)
            ids = [*ids, *[r["id"] for r in search_results]]
        
            # if no ids have been found, return an empty FeatureCollection
            if len(ids) == 0:
                return {"type": "FeatureCollection", "features": []}
        
        # in any other case we go for the locations.
        result = db.get_entries_locations(session, ids=ids, limit=limit, offset=offset)
    
    return result


def licenses(id: int = None, offset: int = None, limit: int = None):
    with connect() as session:
        if id is not None:
            result = db.get_license_by_id(session, id=id)
        else:
            result = db.get_licenses(session, limit=limit, offset=offset)
    
    return result


def authors(id: int = None, entry_id: int = None, search: str = None, exclude_ids: List[int] = None, offset: int = None, limit: int = None) -> List[models.Author]:
    with connect() as session:
        # if an author_id is given, we return only the author of that id
        if id is not None:
            authors = db.get_author_by_id(session, id=id)
        # if an entry_id is given, we return only the authors of that entry
        elif entry_id is not None:
            authors = db.get_authors_by_entry(session, entry_id=entry_id)
        else:
            authors = db.get_authors(session, search=search, exclude_ids=exclude_ids, limit=limit, offset=offset)
    
    return authors


def variables(id: int = None, only_available: bool = False, offset: int = None, limit: int = None) -> List[models.Variable]:
    with connect() as session:
        if only_available:
            variables = db.get_available_variables(session, limit=limit, offset=offset)
        elif id is not None:
            variables = [db.get_variable_by_id(session, id=id)]
        else:
            variables = db.get_variables(session, limit=limit, offset=offset)
    
    return variables

def datatypes(id: int = None) -> List[models.DataSourceType]:
    # TODO: this may need some more parameters
    with connect() as session:
        return db.get_datatypes(session, id=id)


def add_entry(payload: models.MetadataPayload) -> models.Metadata:
    # get the variable
    # TODO: put this into the server, as this is due to the FORM. the core package should use the payload model
    # if 'variable_id' in flat_dict or 'variable.id' in flat_dict:
    #     vid = flat_dict.pop('variable_id', flat_dict.pop('variable.id'))
    #     variable = variables(id=vid)[0]
    
    # # overload the payload with the variable from the database
    # flat_dict['variable'] = variable.model_dump()

    # load the payload model
    #payload = utils.metadata_payload_to_model(flat_dict)

    # add the entry to the database
    with connect() as session:
        metadata = db.add_entry(session, payload=payload)

        # check if there is a datasource in the payload
        if payload.datasource is not None:
            db.add_datasource(session, entry_id=metadata.id, datasource=payload.datasource)
    
    # TODO: this is not ideal as the metadata gets fetched from add_entry already
    final_meta = entries(ids=metadata.id) 
    return final_meta


def add_datasource(entry_id: int, payload: models.DataSource) -> models.DataSource:
    with connect() as session:
        datasource_id = db.add_datasource(session, entry_id=entry_id, datasource=payload)

        
    return datasource_id


from setuptools import setup, find_packages

def version():
    loc = dict()
    with open('metacatalog_api/__version__.py') as f:
        exec(f.read(), loc, loc)
    return loc['__version__']

setup(
    name='metacatalog-api',
    version=version(),
    license='MIT',
    packages=find_packages(),
    install_requires=open('requirements.txt').read().splitlines(),
    author='Mirko Mälicke',
    author_email='mirko.maelicke@kit.edu',
    description='FastAPI backend for managing a Metacatalog instance',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/vforwater/metacatalog-api',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.9',
    include_package_data=True,
)